# bmmp functions v1.3: April 8, 2008: by Iain Pardoe
#
# mmp
# mmp.glm
# mmp.lm
# mmp.default
# bmmp
# bmmp.glm
# bmmp.lm
# bmmp.default
# postsamp.lm
# postsamp.binlog

"mmp" <-
# Construct a Marginal Model Plot for model x, 
# which can be a glm or lm
  function(x, ...)
  {
    UseMethod("mmp")
  }

"mmp.glm" <-
# Support function for mmp that works with glm objects
  function(x, h = NULL, hdefault = NULL, hname, mrange = NULL, ...)
  {

    if(is.null(hdefault)) {

      ### if input h is a single vector ###

      if(!is.null(h) && is.null(dim(h))) {
        name <- deparse(substitute(h))
        h <- cbind(h)
        dimnames(h)[[2]] <- name
      }

      hdefault <- x$linear.predictors
      hname <- "linear fit"

    }
    
    if(x$family[1]=="binomial") {
      mrange <- range(x$y)
    }

    NextMethod("mmp", h = h, hdefault = hdefault, hname = hname,
               mrange = mrange)
    invisible(x)

  }

"mmp.lm" <- 
# Support function for mmp that works with lm objects
  function(x, yname = NULL, h = NULL, hdefault = NULL, hname = NULL,
#           default = T, random = NULL, included = NULL, ...)
           default = T, random = NULL, dmd = F, dmdsp = NULL,
           smoother = "spline", included = NULL, ...)
  {

    ### if default set to F, but h or random not specified ##

    if(default == F && is.null(h) && is.null(random) && ! dmd) {
      stop("You must specify h and/or random/dmd if you set default = F\n")
    }

    ## make sure subsets work, and allow only specified cases to be included ##

    yhat <- fitted.values(x)

    if(is.null(included)) {
      included <- names(yhat)
    }

    else {
      included <- unique(c(names(yhat), included)[duplicated(c(names(yhat),
                                                               included))])
      included <- included[!is.na(as.numeric(included))]
      yhat <- yhat[included]
    }

    yname <- attr(terms(x), "variables")[[attr(terms(x), "response") + 1]]
    resp <- eval(yname)[as.numeric(included)]

    ### add jitter to plotted points if a binary response ###

    respj <-
      if(sort(unique(resp))[1] == 0 && sort(unique(resp))[2] == 1 &&
         length(unique(resp)) == 2) resp + .01 * rnorm(length(resp))
      else resp

    ### if a LM, not a GLM ###

    if(is.null(hdefault)) {
      hdefault <- yhat
      hname <- "fitted values"
    }

    ### if requested, do not include a default direction h ###

    if(default == F) {
      hdefault <- NULL
    }

    ### if h not specified, no directions h other than the default ###

    if(is.null(h)) {
      num <- default + 0
    }

    else {

      ### if input h is a single vector ###

      if(is.null(dim(h))) {
        num <- default + 1
        name <- deparse(substitute(h))
        h <- cbind(h)
        dimnames(h)[[2]] <- name
      }

      ### if input h is at least two vectors ###

      else {
        num <- default + dim(h)[2]
      }

    }

    ### combine hdefault and h ###

    if(!is.null(h)) {
      savehname <- dimnames(h)[[2]]
      h <- as.matrix(h[as.numeric(included), ])
      dimnames(h) <- list(included, savehname)
    }

    if(!is.null(hdefault) || !is.null(h)) {
      h <- cbind(hdefault[included], h)
    }

    ### set name of default direction h if present ###

    if(default) {
      dimnames(h)[[2]][1] <- hname
    }

    ### generate random directions h if "random" specified

    if(!is.null(random)) {

      xmatrix <-
        if(attr(terms(x), "intercept"))
          as.matrix(model.matrix(x)[included, - 1])
        else model.matrix(x)[included, ]
      randv <- rnorm(dim(xmatrix)[2])
      randh <- scale(xmatrix) %*% randv / sqrt(sum(randv ^ 2))
      dimnames(randh)[[2]] <- paste("random direction", 1)

      if(random > 1) {
        for (i in 1:(random - 1)) {
          randv <- rnorm(dim(xmatrix)[2])
          rand <- scale(xmatrix) %*% randv / sqrt(sum(randv ^ 2))
          randh <- cbind(randh, rand)
          dimnames(randh)[[2]][i + 1] <- paste("random direction", i + 1)
        }

      }

      h <- cbind(h, randh)
      num <- num + random

    }

    ### if weighted model ###

    w <- eval(match.call(function(weights, ...){},
                         x$call)$weights)[as.numeric(included)]

    ### if unweighted model ###

    if(is.null(w)) {
      w <- rep(1, length(resp))
    }

    ### generate discrepant marginals direction h if "dmd" specified
    
    if(dmd) {

      if(is.null(dmdsp)) dmdsp <- switch(smoother,
                                         spline = 4,
                                         loess1 = .7,
                                         loess2 = .9)
      cat("Calculate discrepant marginals direction with smoothing parameter",
          dmdsp, "...\n")
      xmatrix <-
        if(attr(terms(x), "intercept"))
          scale(as.matrix(model.matrix(x)[included, - 1]))
        else scale(model.matrix(x)[included, ])
      numevals <- 50
      theta <- seq(0, pi, length = numevals)
      oldh <- xmatrix[, 1]

      for(i in 1:(dim(xmatrix)[2] - 1)) {
        candh <- t(cos(theta) %o% as.vector(oldh) + sin(theta) %o%
                   xmatrix[, i + 1])
        ymod <- switch(smoother,
                       spline = apply(candh, 2, function(x)
                         predict(smooth.spline(x, y = yhat, w = w,
                                               df = dmdsp), x = x)$y),
                       loess1 = apply(candh, 2, function(x)
                         fitted.values(loess(yhat ~ x, weights = w,
                                             span = dmdsp, degree = 1,
                                             trace.hat = "approximate"))),
                       loess2 = apply(candh, 2, function(x)
                         fitted.values(loess(yhat ~ x, weights = w,
                                             span = dmdsp, degree = 2,
                                             trace.hat = "approximate"))))
        ydat <- switch(smoother,
                       spline = apply(candh, 2, function(x)
                         predict(smooth.spline(x, y = respj, w = w, df = 4),
                                 x = x)$y),
                       loess1 = apply(candh, 2, function(x)
                         fitted.values(loess(respj ~ x, weights = w, span = .7,
                                             degree = 1,
                                             trace.hat = "approximate"))),
                       loess2 = apply(candh, 2, function(x)
                         fitted.values(loess(respj ~ x, weights = w, span = .9,
                                             degree = 2,
                                             trace.hat = "approximate"))))
        dist <- apply((ymod - ydat)^2, 2, sum)
        oldh <- scale(candh[, which.max(dist)])
#        plot(theta, dist)
        cat("h", i, " is ",
            round(cos(theta[which.max(dist)]), 3), " scale(h", i - 1, ") + ",
            round(sin(theta[which.max(dist)]), 3), " scale(x", i + 1, ")\n",
            sep = "")
      }

      dmdh <- oldh
      dimnames(dmdh)[[2]] <- paste("discrepant marginals direction")

      ###      h1 is ch0     * scale(h0)     + cx2     * scale(x2)   [h0 = x1]
      ###      h2 is ch1     * scale(h1)     + cx3     * scale(x3)
      ###      h3 is ch2     * scale(h2)     + cx4     * scale(x4)
      ### ... 
      ###  h(p-2) is ch(p-3) * scale(h(p-3)) + cx(p-1) * scale(x(p-1))
      ###   dmd h is ch(p-2) * scale(h(p-2)) + cxp     * scale(xp)
      
      h <- cbind(h, dmdh)
      num <- num + 1

    }

    ### if GLM ###

    disp <- summary(x)$dispersion 

    ### if LM ###

    if(is.null(disp) || family(x)$family[1]=="Gamma") {
      disp <- deviance(x) / x$df.residual
    }

    varmod <- family(x)$variance(fitted.values(x))
    if(length(varmod) > 1) {
      names(varmod) <- dimnames(model.matrix(x))[[1]]
      varmod <- varmod[included]
    }
    varmod <- disp * varmod / w

    NextMethod("mmp", resp = resp, respj = respj, yname = yname, yhat = yhat,
#               h = h, num = num, w = w, varmod = varmod, ...)
               h = h, num = num, w = w, varmod = varmod, smoother = smoother,
               ...)
    invisible(x)

  }

"mmp.default" <- 
# Support function for mmp that does the final calculations and plotting
  function(x, resp, respj, yname, yhat, h, num, w, varmod, incvar = T,
#           adjvar = F, smoother = "spline", sparams = NULL, spar = NULL,
           adjvar = F, smoother, sparams = NULL, spar = NULL,
           spdefault = F, mrange = NULL, mult = .05, sptry = NULL,
#           spdefault = F, cv = T, mrange = NULL, mult = .05, sptry = NULL,
           spartry = NULL, usetrellis = F, ...)
  {

    old.par <- par(no.readonly = TRUE, pty = "s")
    on.exit(par(old.par))
    old.options <- options(warn = -1) # suppress warnings
    on.exit(options(old.options))
    frame()

    smooth <- switch(smoother,
                     spline = function(xaxis, yaxis, wts, sp, spalt)
                     smooth.spline(x = xaxis, y = yaxis, w = wts, df = sp,
                                   spar = spalt),
                     loess1 = function(xaxis, yaxis, wts, sp, ...)
                     loess(yaxis ~ xaxis, weights = wts, span = sp, degree = 1,
                           trace.hat = "approximate"),
                     loess2 = function(xaxis, yaxis, wts, sp, ...)
                     loess(yaxis ~ xaxis, weights = wts, span = sp, degree = 2,
                           trace.hat = "approximate"),
                     kernel = function(xaxis, yaxis, wts, sp, ...)
                     ksmooth(x = xaxis, y = yaxis, kernel = "normal",
                             x.points = xaxis,
                             bandwidth = diff(range(xaxis)) * sp),
                     supsmu = function(xaxis, yaxis, wts, sp, ...)
                     supsmu(x = xaxis, y = yaxis, wt = wts, span = sp),
                     stop(paste("Smoother ", smoother, " not defined\n",
                               "Try spline, loess1, loess2, kernel, or supsmu",
                                sep = ""),
                          call. = FALSE))
    
    if(smoother == "spline" || smoother == "kernel")
      smtype <- 1 # output x in increasing order, duplicates retained

    if(smoother == "supsmu")
      smtype <- 2 # output x in increasing order, duplicates removed

    if(smoother == "loess1" || smoother == "loess2")
      smtype <- 3 # output x in original input x order

    if(smtype == 2 && incvar) 
      cat("Warning: variance smooths not available for this smoother\n")

    if(sort(unique(resp))[1] == 0 && sort(unique(resp))[2] == 1 &&
       length(unique(resp)) == 2 && incvar) {
      cat("Warning: variance smooths not needed for binary logistic\n")
      incvar <- F
    }

    ### set up smoothing parameter defaults if needed ###

    if(spdefault) {

      if(is.null(sparams) && is.null(spar)) {
        sparams <- switch(smoother,
                          spline = rep(4, num),
                          loess1 = rep(.7, num),
                          loess2 = rep(.9, num),
                          kernel = rep(.3, num),
                          supsmu = rep(.5, num))
      }

      else {
        stop("You cannot specify sparams or spar and set spdefault = T\n",
             call. = FALSE)
      }
  
    }

    ### set up smoothing parameters to try if needed ###

    if(is.null(sparams) && is.null(spar)) {
      if(is.null(sptry)) {
        if(is.null(spartry)) {
          sptry <- switch(smoother,
                          spline = c(3, 4, 5, 6),
                          loess1 = c(.8, .7, .6, .5),
                          loess2 = c(1, .9, .8, .7),
                          kernel = c(.4, .3, .2, .1),
                          supsmu = c(.6, .5, .4, .3))
        }
        else sptry <- rep(0, length(spartry))
      }
      spartry <-
#        if(smoother == "spline" && is.null(spartry)) rep(0, length(sptry))
        if(smoother == "spline" && is.null(spartry)) NULL
        else spartry
    }

    ### trellis graphics ###

    if(usetrellis) {
      stop("Trellis graphics not available for R\n", call. = FALSE)
    }

    ### traditional graphics ###

    else {
  
      ### interactively select smoothing parameters ###

      if(is.null(sparams) && is.null(spar)) {

        ### cycle through number of directions h ###
    
        for(i in 1:num) {

#          if(cv) {
#            sparams[i] <- max(smooth.spline(h[, i], yhat, cv = T)$df,
#                              smooth.spline(h[, i], respj, cv = T)$df)
#            sparams[i] <- smooth.spline(h[, i], yhat, cv = T)$df
#            sparams[i] <- smooth.spline(h[, i], respj, cv = T)$df
#          }

#          else {

          close.screen(all = T)
          split.screen(figs = c(1, 2))

          ### left screen for fitted values ###      

          screen(1)
          plot(h[, i], yhat, xlim = range(h[, i]),
               ylim = c(min(resp) - mult * diff(range(resp)),
                 max(resp) + mult * diff(range(resp))),
               xlab = paste("h = ", dimnames(h)[[2]][i]),
               ylab = "Fitted values (under M)", xpd = F)

          ### cycle through number of smoothing parameters to try ###
      
          for(j in seq(along = sptry)) {

            mmodel <- smooth(xaxis = h[, i], yaxis = yhat, wts = w,
                             sp = sptry[j], spalt = spartry[j])

            if(smtype == 1 || smtype == 2) {
              plotx <- mmodel$x
              mmodely <- mmodel$y
            }

            if(smtype == 3) {
              plotx <- sort(h[, i])
              mmodely <- fitted.values(mmodel)[order(h[, i])]
            }

            ### when there is a restriction on range for y, eg in (0,1) ###

            if(!is.null(mrange)) {
              mmodely <- pmin(pmax(mmodely, mrange[1]), mrange[2])
            }

            ### add line for model to plot ###
        
            lines(plotx, mmodely, lty = 1, lwd = 2, col = j + 1)

          }

          ### right screen for data values ###
      
          screen(2)
          plot(h[, i], respj, xlim = range(h[, i]),
               ylim = c(min(resp) - mult * diff(range(resp)),
                 max(resp) + mult * diff(range(resp))),
               xlab = paste("h = ", dimnames(h)[[2]][i]),
               ylab = deparse(yname), xpd = F)

          ### cycle through number of smoothing parameters to try ###
      
          for(j in seq(along = sptry)) {

            mdata <- smooth(xaxis = h[, i], yaxis = resp, wts = w,
                             sp = sptry[j], spalt = spartry[j])

            if(smtype == 1 || smtype == 2) {
              plotx <- mdata$x
              mdatay <- mdata$y
            }

            if(smtype == 3) {
              mdatay <- fitted.values(mdata)[order(h[, i])]
            }

            ### when there is a restriction on range for y, eg in (0,1) ###
        
            if(!is.null(mrange)) {
              mdatay <- pmin(pmax(mdatay, mrange[1]), mrange[2])
            }

            ### add line for data to plot ###

            lines(plotx, mdatay, lwd = 2, col = j + 1)

          }

          ### legend for plot ###
      
          close.screen(all = T)
          par(new = TRUE)
          plot(0, 0, axes = F, xlab = "", ylab = "",
               main = paste(smoother, "smoothing parameters:"))
          legend(x = 0, y = 1,
                 horiz = TRUE, xjust = .5, x.intersp = .5,
                 legend = as.character(if(smoother == "spline" &&
#                   spartry[1] > 0) spartry else sptry),
                   !is.null(spartry)) spartry else sptry),
                 lty = 1, lwd = 2, col = seq(2, along = sptry))

          ### user prompted for best smoothing parameter ###

          choice <-
#            if(smoother == "spline" && spartry[1] > 0) spartry
            if(smoother == "spline" && !is.null(spartry)) spartry
            else sptry
          pick <- menu(choice,
                       title = "Select best smoothing parameter (or 0 to exit)"
                       )

          if(pick) {
            sparams[i] <- choice[pick]
          }
      
          else stop("No action\n", call. = FALSE)

          if(smoother == "spline") {

#            if(spartry[1] > 0) {
            if(!is.null(spartry)) {
              spar[i] <- sparams[i]
              sparams[i] <- sptry[1]
            }

#            else spar[i] <- 0
            else spar[i] <- NULL

          }

        }

#        }

      }

      if(smoother == "spline") {

        if(is.null(sparams)) {
          if(length(spar) != num)
            stop("spar must have the same number of components as h\n",
                 call. = FALSE)
          sparams <- rep(2, length(spar))
        }

        if(is.null(spar)) {
          if(length(sparams) != num)
            stop("sparams must have the same number of components as h\n",
                 call. = FALSE)
#          spar <- rep(0, length(sparams))
          spar <- NULL
        }

      }

      else if(length(sparams) != num) {
        stop("sparams must have the same number of components as h\n",
             call. = FALSE)
      }
      
      ### plot mmp's; divide plotting region according to num ###

      close.screen(all = T)
      split.screen(figs = c(round(sqrt(num), 0), ceiling(sqrt(num))))

      ### cycle through number of directions h ###
    
      for(i in 1:num) {

        mmodel <- smooth(xaxis = h[, i], yaxis = yhat, wts = w,
                         sp = sparams[i], spalt = spar[i])

        mdata <- smooth(xaxis = h[, i], yaxis = resp, wts = w,
                        sp = sparams[i], spalt = spar[i])

        if(smtype == 1 || smtype == 2) {
          plotx <- mmodel$x
          mmodely <- mmodel$y
          mdatay <- mdata$y
        }

        if(smtype == 3) {
          plotx <- sort(h[, i])
          mmodely <- fitted.values(mmodel)[order(h[, i])]
          mdatay <- fitted.values(mdata)[order(h[, i])]
        }

        ### when there is a restriction on range for y, eg in (0,1) ###

        if(!is.null(mrange)) {
          mmodely <- pmin(pmax(mmodely, mrange[1]), mrange[2])
          mdatay <- pmin(pmax(mdatay, mrange[1]), mrange[2])
        }

        ### construct the mmp ###

        screen(i)
        par(mex = 1 - (num - 1) * .05, cex = 1 - (num - 1) * .05)
        plot(h[, i], respj, xlim = range(h[, i]),
             ylim = c(min(resp) - mult * diff(range(resp)),
               max(resp) + mult * diff(range(resp))),
             xlab = paste(dimnames(h)[[2]][i], " (sp:",
#               if(smoother == "spline" && spar[1] > 0) spar[i] else sparams[i],
               if(smoother == "spline" && !is.null(spar)) spar[i] else sparams[i],
               ")", sep=""),
             ylab = deparse(yname), xpd = F)

        ### add lines to plot ###

        lines(plotx, mmodely, lty = 1, lwd = 2, col = "red")
        lines(plotx, mdatay, lwd = 2, col = "blue")

        ### calculate variance smooths if they are wanted as well ###
        ### not available for "supsmu" since it removes duplicate x-values ###
    
        if(smtype == 1 || smtype == 3) {

          if(incvar) {

            ## bias adjustment as in Ruppert, Wand, Holst and Hossjer (1997) ##
            ## only implemented for "spline", and very slow ##

            if(adjvar && smoother == "spline") {
              lx <- length(plotx)
              adjarray <- array(c(rep(plotx, lx), diag(lx)), c(lx, lx, 2))
              adjsmooth <- apply(adjarray, 2, smooth.spline, df = sparams[i])
              smatrix <- sapply(adjsmooth, function(x) x$y)
              del <- diag(smatrix %*% t(smatrix) - 2 * smatrix)
            }
      
            else {
              del <- 0
            }

            evmodel <- smooth(xaxis = h[, i], yaxis = varmod, wts = w,
                              sp = sparams[i], spalt = spar[i])

            sqresmodel <- switch(smoother,
                                 spline = (yhat - predict(mmodel, h[, i])$y)^2,
                                 loess1 = (yhat - fitted.values(mmodel))^2,
                                 loess2 = (yhat - fitted.values(mmodel))^2,
                                 kernel = (yhat[order(h[, i])] - mmodely)^2)

            vemodel <- smooth(xaxis = h[, i], yaxis = sqresmodel, wts = w,
                              sp = sparams[i], spalt = spar[i])

            if(smtype == 1) {
              vmodely <- evmodel$y + vemodel$y
            }
      
            if(smtype == 3) {
              vmodely <- (fitted.values(evmodel) +
                          fitted.values(vemodel))[order(h[, i])]
            }

            smodely <- sqrt(ifelse(vmodely <= 0, 0, vmodely))

            sqresdata <- switch(smoother,
                                spline = (resp - predict(mdata, h[, i])$y)^2,
                                loess1 = (resp - fitted.values(mdata))^2,
                                loess2 = (resp - fitted.values(mdata))^2,
                                kernel = (resp[order(h[, i])] - mdatay)^2)

            allsqresdata <- cbind(if(exists("allsqresdata")) allsqresdata
                                  else NULL,
                                  sqresdata)

            vdata <- smooth(xaxis = h[, i], yaxis = sqresdata, wts = w,
                            sp = sparams[i], spalt = spar[i])

            if(smtype == 1) {
              vdatay <- vdata$y
            }
      
            if(smtype == 3) {
              vdatay <- (fitted.values(vdata))[order(h[, i])]
            }

            sdatay <- sqrt(ifelse(vdatay <= 0, 0, vdatay))

            lines(plotx, mmodely - smodely, lty = 1, lwd = 1, col = "red")
            lines(plotx, mmodely + smodely, lty = 1, lwd = 1, col = "red")
            lines(plotx, mdatay - sdatay, lwd = 1, col = "blue")
            lines(plotx, mdatay + sdatay, lwd = 1, col = "blue")

          }

        }

        ### legend for plot ###

        if(i == num) {
          close.screen(all = T)
          par(new = TRUE, mar = c(5.1, 2.1, 1.1, 2.1))
          plot(0, 0, axes = F, xlab = "", ylab = "", type = "n",
               main = paste("Marginal model plot(s):", smoother, "smooths"))
          legend(x = 0, y = 1.05,
                 horiz = TRUE, xjust = .5, x.intersp = .5, cex = .85,
                 legend = c("Estimates for data (under F)", 
                   "Estimates for fitted values (under M)"),
                 lty = c(1, 1), lwd = 2, col = c("blue", "red"))
          par(mar = c(5.1, 4.1, 4.1, 2.1)) # don't know why this is needed
        }

      }

      close.screen(all = T)
      invisible()

    }

  }

"bmmp" <-
# Construct a Bayes Marginal Model Plot for model x, 
# which can be a glm or lm
  function(x, ...)
  {
    UseMethod("bmmp")
  }

"bmmp.glm" <-
# Support function for bmmp that works with glm objects
  function(x, h = NULL, hdefault = NULL, hname, mrange = NULL, ...)
  {

    if(is.null(hdefault)) {

      ### if input h is a single vector ###

      if(!is.null(h) && is.null(dim(h))) {
        name <- deparse(substitute(h))
        h <- cbind(h)
        dimnames(h)[[2]] <- name
      }

      hdefault <- x$linear.predictors
      hname <- "linear fit"

    }

    if(x$family[1]=="binomial") {
      mrange <- range(x$y)
    }

    NextMethod("bmmp", h = h, hdefault = hdefault, hname = hname,
               mrange = mrange)
    invisible(x)

  }

"bmmp.lm" <- 
# Support function for bmmp that works with lm objects
  function(x, yname = NULL, h = NULL, hdefault = NULL, hname = NULL,
           default = T, random = NULL, dmd = F, dmdsp = NULL,
           smoother = "spline", included = NULL, ...)
  {

    ### if default set to F, but h or random not specified ##

    if(default == F && is.null(h) && is.null(random) && ! dmd) {
      stop("You must specify h and/or random/dmd if you set default = F\n")
    }

    ## make sure subsets work, and allow only specified cases to be included ##

    yhat <- fitted.values(x)

    if(is.null(included)) {
      included <- names(yhat)
    }

    else {
      included <- unique(c(names(yhat), included)[duplicated(c(names(yhat),
                                                               included))])
      included <- included[!is.na(as.numeric(included))]
      yhat <- yhat[included]
    }

    yname <- attr(terms(x), "variables")[[attr(terms(x), "response") + 1]]
    resp <- eval(yname)[as.numeric(included)]

    ### add jitter to plotted points if a binary response ###

    respj <-
      if(sort(unique(resp))[1] == 0 && sort(unique(resp))[2] == 1 &&
         length(unique(resp)) == 2) resp + .01 * rnorm(length(resp))
      else resp

    ### if a LM, not a GLM ###

    if(is.null(hdefault)) {
      hdefault <- yhat
      hname <- "fitted values"
    }

    ### if requested, do not include a default direction h ###

    if(default == F) {
      hdefault <- NULL
    }

    ### if h not specified, no directions h other than the default ###

    if(is.null(h)) {
      num <- default + 0
    }

    else {

      ### if input h is a single vector ###

      if(is.null(dim(h))) {
        num <- default + 1
        name <- deparse(substitute(h))
        h <- cbind(h)
        dimnames(h)[[2]] <- name
      }

      ### if input h is at least two vectors ###

      else {
        num <- default + dim(h)[2]
      }

    }

    ### combine hdefault and h ###

    if(!is.null(h)) {
      savehname <- dimnames(h)[[2]]
      h <- as.matrix(h[as.numeric(included), ])
      dimnames(h) <- list(included, savehname)
    }

    if(!is.null(hdefault) || !is.null(h)) {
      h <- cbind(hdefault[included], h)
    }

    ### set name of default direction h if present ###

    if(default) {
      dimnames(h)[[2]][1] <- hname
    }

    ### generate random directions h if "random" specified

    if(!is.null(random)) {

      xmatrix <-
        if(attr(terms(x), "intercept"))
          as.matrix(model.matrix(x)[included, - 1])
        else model.matrix(x)[included, ]
      randv <- rnorm(dim(xmatrix)[2])
      randh <- scale(xmatrix) %*% randv / sqrt(sum(randv ^ 2))
      dimnames(randh)[[2]] <- paste("random direction", 1)

      if(random > 1) {
        for (i in 1:(random - 1)) {
          randv <- rnorm(dim(xmatrix)[2])
          rand <- scale(xmatrix) %*% randv / sqrt(sum(randv ^ 2))
          randh <- cbind(randh, rand)
          dimnames(randh)[[2]][i + 1] <- paste("random direction", i + 1)
        }

      }

      h <- cbind(h, randh)
      num <- num + random

    }

    ### if weighted model ###

    w <- eval(match.call(function(weights, ...){},
                         x$call)$weights)[as.numeric(included)]

    ### if unweighted model ###

    if(is.null(w)) {
      w <- rep(1, length(resp))
    }

    ### generate discrepant marginals direction h if "dmd" specified
    
    if(dmd) {

      if(is.null(dmdsp)) dmdsp <- switch(smoother,
                                         spline = 4,
                                         loess1 = .7,
                                         loess2 = .9)
#      cat("Calculate discrepant marginals direction with smoothing parameter",
#          dmdsp, "...\n")
      xmatrix <-
        if(attr(terms(x), "intercept"))
          scale(as.matrix(model.matrix(x)[included, - 1]))
        else scale(model.matrix(x)[included, ])
      numevals <- 50
      theta <- seq(0, pi, length = numevals)
      oldh <- xmatrix[, 1]

      for(i in 1:(dim(xmatrix)[2] - 1)) {
        candh <- t(cos(theta) %o% as.vector(oldh) + sin(theta) %o%
                   xmatrix[, i + 1])
        ymod <- switch(smoother,
                       spline = apply(candh, 2, function(x)
                         predict(smooth.spline(x, y = yhat, w = w,
                                               df = dmdsp), x = x)$y),
                       loess1 = apply(candh, 2, function(x)
                         fitted.values(loess(yhat ~ x, weights = w,
                                             span = dmdsp, degree = 1,
                                             trace.hat = "approximate"))),
                       loess2 = apply(candh, 2, function(x)
                         fitted.values(loess(yhat ~ x, weights = w,
                                             span = dmdsp, degree = 2,
                                             trace.hat = "approximate"))))
        ydat <- switch(smoother,
                       spline = apply(candh, 2, function(x)
                         predict(smooth.spline(x, y = respj, w = w, df = 4),
                                 x = x)$y),
                       loess1 = apply(candh, 2, function(x)
                         fitted.values(loess(respj ~ x, weights = w, span = .7,
                                             degree = 1,
                                             trace.hat = "approximate"))),
                       loess2 = apply(candh, 2, function(x)
                         fitted.values(loess(respj ~ x, weights = w, span = .9,
                                             degree = 2,
                                             trace.hat = "approximate"))))
        dist <- apply((ymod - ydat)^2, 2, sum)
        oldh <- scale(candh[, which.max(dist)])
#        plot(theta, dist)
#        cat("h", i, " is ",
#            round(cos(theta[which.max(dist)]), 3), " scale(h", i - 1, ") + ",
#            round(sin(theta[which.max(dist)]), 3), " scale(x", i + 1, ")\n",
#            sep = "")
      }

      dmdh <- oldh
#      dimnames(dmdh)[[2]] <- paste("discrepant marginals direction")
      dimnames(dmdh)[[2]] <- paste("dmd")

      ###      h1 is ch0     * scale(h0)     + cx2     * scale(x2)   [h0 = x1]
      ###      h2 is ch1     * scale(h1)     + cx3     * scale(x3)
      ###      h3 is ch2     * scale(h2)     + cx4     * scale(x4)
      ### ... 
      ###  h(p-2) is ch(p-3) * scale(h(p-3)) + cx(p-1) * scale(x(p-1))
      ###   dmd h is ch(p-2) * scale(h(p-2)) + cxp     * scale(xp)
      
      h <- cbind(h, dmdh)
      num <- num + 1

    }

    ### if GLM ###

    disp <- summary(x)$dispersion 

    ### if LM ###

    if(is.null(disp) || family(x)$family[1]=="Gamma") {
      disp <- deviance(x) / x$df.residual
    }

    varmod <- family(x)$variance(fitted.values(x))
    if(length(varmod) > 1) {
      names(varmod) <- dimnames(model.matrix(x))[[1]]
      varmod <- varmod[included]
    }
    varmod <- disp * varmod / w

    NextMethod("bmmp", resp = resp, respj = respj, yname = yname, yhat = yhat,
               h = h, num = num, w = w, varmod = varmod, included = included,
#               ...)
               smoother = smoother, ...)
    invisible(x)

  }

"bmmp.default" <- 
# Support function for bmmp that does the final calculations and plotting
  function(x, resp, respj, yname, yhat, samp, h, num, w, varmod, included,
#           incvar = T, adjvar = F, smoother = "spline", sparams = NULL,
           incvar = T, adjvar = F, smoother, sparams = NULL,
           spar = NULL, spdefault = F, mrange = NULL, mult = .15, sptry = NULL,
           spartry = NULL, usetrellis = F, bdm = T, incplots = T, ...)
  {

    old.par <- par(no.readonly = TRUE, pty = "s")
#    old.par <- par(no.readonly = TRUE, pty = "s", omi = c(0, 0, 0, 0))
    on.exit(par(old.par))
    old.options <- options(warn = -1) # suppress warnings
    on.exit(options(old.options))
    frame()

    smooth <- switch(smoother,
                     spline = function(xaxis, yaxis, wts, sp, spalt)
                     smooth.spline(x = xaxis, y = yaxis, w = wts, df = sp,
                                   spar = spalt),
                     loess1 = function(xaxis, yaxis, wts, sp, ...)
                     loess(yaxis ~ xaxis, weights = wts, span = sp, degree = 1,
                           trace.hat = "approximate"),
                     loess2 = function(xaxis, yaxis, wts, sp, ...)
                     loess(yaxis ~ xaxis, weights = wts, span = sp, degree = 2,
                           trace.hat = "approximate"),
                     kernel = function(xaxis, yaxis, wts, sp, ...)
                     ksmooth(x = xaxis, y = yaxis, kernel = "normal",
                             x.points = xaxis,
                             bandwidth = diff(range(xaxis)) * sp),
                     supsmu = function(xaxis, yaxis, wts, sp, ...)
                     supsmu(x = xaxis, y = yaxis, wt = wts, span = sp),
                     stop(paste("Smoother ", smoother, " not defined\n",
                               "Try spline, loess1, loess2, kernel, or supsmu",
                                sep = ""),
                          call. = FALSE))
    
    if(smoother == "spline" || smoother == "kernel")
      smtype <- 1 # output x in increasing order, duplicates retained

    if(smoother == "supsmu")
      smtype <- 2 # output x in increasing order, duplicates removed

    if(smoother == "loess1" || smoother == "loess2")
      smtype <- 3 # output x in original input x order

    if(smtype == 2 && incvar) 
      cat("Warning: variance smooths not available for this smoother\n")

    if(smtype == 2 && bdm) 
      cat("Warning: BDM calculation not available for this smoother\n")

    if(sort(unique(resp))[1] == 0 && sort(unique(resp))[2] == 1 &&
       length(unique(resp)) == 2 && incvar) {
      cat("Warning: variance smooths not needed for binary logistic\n")
      incvar <- F
    }

    ### set up smoothing parameter defaults if needed ###

    if(spdefault) {

      if(is.null(sparams) && is.null(spar)) {
        sparams <- switch(smoother,
                          spline = rep(4, num),
                          loess1 = rep(.7, num),
                          loess2 = rep(.9, num),
                          kernel = rep(.3, num),
                          supsmu = rep(.5, num))
      }

      else {
        stop("You cannot specify sparams or spar and set spdefault = T\n",
             call. = FALSE)
      }

    }
  
    ### set up smoothing parameters to try if needed ###

    if(is.null(sparams) && is.null(spar)) {
      if(is.null(sptry)) {
        if(is.null(spartry)) {
          sptry <- switch(smoother,
                          spline = c(3, 4, 5, 6),
                          loess1 = c(.8, .7, .6, .5),
                          loess2 = c(1, .9, .8, .7),
                          kernel = c(.4, .3, .2, .1),
                          supsmu = c(.6, .5, .4, .3))
        }
        else sptry <- rep(0, length(spartry))
      }
      spartry <-
#        if(smoother == "spline" && is.null(spartry)) rep(0, length(sptry))
        if(smoother == "spline" && is.null(spartry)) NULL
        else spartry
    }

    ### trellis graphics ###

    if(usetrellis) {
      stop("Trellis graphics not available for R\n", call. = FALSE)
    }

    ### traditional graphics ###

    else {

      ### interactively select smoothing parameters ###
  
      if(is.null(sparams) && is.null(spar)) {

        ### cycle through number of directions h ###
    
        for(i in 1:num) {

          close.screen(all = T)
          split.screen(figs = c(1, 2))

          ### left screen for fitted values ###      

          screen(1)
          plot(h[, i], yhat, xlim = range(h[, i]),
               ylim = c(min(resp) - mult * diff(range(resp)),
                 max(resp) + mult * diff(range(resp))),
               xlab = paste("h = ", dimnames(h)[[2]][i]),
               ylab = "Fitted values (under M)", xpd = F)

          ### cycle through number of smoothing parameters to try ###

          for(j in seq(along = sptry)) {

            mmodel <- smooth(xaxis = h[, i], yaxis = yhat, wts = w,
                             sp = sptry[j], spalt = spartry[j])

            if(smtype == 1 || smtype == 2) {
              plotx <- mmodel$x
              mmodely <- mmodel$y
            }

            if(smtype == 3) {
              plotx <- sort(h[, i])
              mmodely <- fitted.values(mmodel)[order(h[, i])]
            }

            ### when there is a restriction on range for y, eg in (0,1) ###

            if(!is.null(mrange)) {
              mmodely <- pmin(pmax(mmodely, mrange[1]), mrange[2])
            }

            ### add line for model to plot ###

            lines(plotx, mmodely, lty = 1, lwd = 2, col = j + 1)

          }

          ### right screen for data values ###
      
          screen(2)
          plot(h[, i], respj, xlim = range(h[, i]),
               ylim = c(min(resp) - mult * diff(range(resp)),
                 max(resp) + mult * diff(range(resp))),
               xlab = paste("h = ", dimnames(h)[[2]][i]),
               ylab = deparse(yname), xpd = F)

          ### cycle through number of smoothing parameters to try ###
      
          for(j in seq(along = sptry)) {

            mdata <- smooth(xaxis = h[, i], yaxis = resp, wts = w,
                             sp = sptry[j], spalt = spartry[j])

            if(smtype == 1 || smtype == 2) {
              plotx <- mdata$x
              mdatay <- mdata$y
            }

            if(smtype == 3) {
              mdatay <- fitted.values(mdata)[order(h[, i])]
            }

            ### when there is a restriction on range for y, eg in (0,1) ###
        
            if(!is.null(mrange)) {
              mdatay <- pmin(pmax(mdatay, mrange[1]), mrange[2])
            }

            ### add line for data to plot ###

            lines(plotx, mdatay, lwd = 2, col = j + 1)

          }

          ### legend for plot ###
      
          close.screen(all = T)
          par(new = TRUE)
          plot(0, 0, axes = F, xlab = "", ylab = "",
               main = paste(smoother, "smoothing parameters:"))
          legend(x = 0, y = 1,
                 horiz = TRUE, xjust = .5, x.intersp = .5,
                 legend = as.character(if(smoother == "spline" &&
#                   spartry[1] > 0) spartry else sptry),
                   !is.null(spartry)) spartry else sptry),
                 lty = 1, lwd = 2, col = seq(2, along = sptry))

          ### user prompted for best smoothing parameter ###

#          choice <- if(smoother == "spline" && spartry[1] > 0) spartry
          choice <- if(smoother == "spline" && !is.null(spartry)) spartry
                    else sptry
          pick <- menu(choice,
                       title = "Select best smoothing parameter (or 0 to exit)"
                       )

          if(pick) {
            sparams[i] <- choice[pick]
          }
      
          else stop("No action\n", call. = FALSE)

          if(smoother == "spline") {

#            if(spartry[1] > 0) {
            if(!is.null(spartry)) {
              spar[i] <- sparams[i]
              sparams[i] <- sptry[1]
            }

#            else spar[i] <- 0
            else spar[i] <- NULL

          }

        }

      }

      if(smoother == "spline") {

        if(is.null(sparams)) {
          if(length(spar) != num)
            stop("spar must have the same number of components as h\n",
                 call. = FALSE)
          sparams <- rep(2, length(spar))
        }
        
        if(is.null(spar)) {
          if(length(sparams) != num)
            stop("sparams must have the same number of components as h\n",
                 call. = FALSE)
#          spar <- rep(0, length(sparams))
          spar <- NULL
        }
        
      }
      
      else if(length(sparams) != num) {
        stop("sparams must have the same number of components as h\n",
             call. = FALSE)
      }

      ### plot mmp's and bmmp's; divide plotting region according to num ###

      close.screen(all = T)
      split.screen(figs = c(round(sqrt(num), 0), ceiling(sqrt(num))))

      ### cycle through number of directions h ###

      if(incplots) {
      for(i in 1:num) {

        mmodel <- smooth(xaxis = h[, i], yaxis = yhat, wts = w,
                         sp = sparams[i], spalt = spar[i])

        mdata <- smooth(xaxis = h[, i], yaxis = resp, wts = w,
                        sp = sparams[i], spalt = spar[i])

        if(smtype == 1 || smtype == 2) {
          plotx <- mmodel$x
          mmodely <- mmodel$y
          mdatay <- mdata$y
        }

        if(smtype == 3) {
          plotx <- sort(h[, i])
          mmodely <- fitted.values(mmodel)[order(h[, i])]
          mdatay <- fitted.values(mdata)[order(h[, i])]
        }

        ### when there is a restriction on range for y, eg in (0,1) ###

        if(!is.null(mrange)) {
          mmodely <- pmin(pmax(mmodely, mrange[1]), mrange[2])
          mdatay <- pmin(pmax(mdatay, mrange[1]), mrange[2])
        }

        ### construct the mmp ###

        screen(i)
        par(mex = 1 - (num - 1) * .05, cex = 1 - (num - 1) * .05)
        plot(h[, i], respj, xlim = range(h[, i]),
             ylim = c(min(resp) - mult * diff(range(resp)),
               max(resp) + mult * diff(range(resp))),
             xlab = paste(dimnames(h)[[2]][i], " (sp:",
#               if(smoother == "spline" && spar[1] > 0) spar[i] else sparams[i],
               if(smoother == "spline" && !is.null(spar)) spar[i] else sparams[i],
               ")", sep=""),
             ylab = deparse(yname), xpd = F)
    
        ### add lines to plot ###

        lines(plotx, mmodely, lty = 1, lwd = 2, col = "red")
        lines(plotx, mdatay, lwd = 2, col = "blue")

        ### calculate variance smooths if they are wanted as well ###
        ### not available for "supsmu" since it removes duplicate x-values ###

        if(smtype == 1 || smtype == 3) {

          if(incvar) {

            evmodel <- smooth(xaxis = h[, i], yaxis = varmod, wts = w,
                              sp = sparams[i], spalt = spar[i])

            sqresmodel <- switch(smoother,
                                 spline = (yhat - predict(mmodel, h[, i])$y)^2,
                                 loess1 = (yhat - fitted.values(mmodel))^2,
                                 loess2 = (yhat - fitted.values(mmodel))^2,
                                 kernel = (yhat[order(h[, i])] - mmodely)^2)

            vemodel <- smooth(xaxis = h[, i], yaxis = sqresmodel, wts = w,
                              sp = sparams[i], spalt = spar[i])

            if(smtype == 1) {
              vmodely <- evmodel$y + vemodel$y
            }
      
            if(smtype == 3) {
              vmodely <- (fitted.values(evmodel) +
                          fitted.values(vemodel))[order(h[, i])]
            }

            smodely <- sqrt(ifelse(vmodely <= 0, 0, vmodely))      

            sqresdata <- switch(smoother,
                                spline = (resp - predict(mdata, h[, i])$y)^2,
                                loess1 = (resp - fitted.values(mdata))^2,
                                loess2 = (resp - fitted.values(mdata))^2,
                                kernel = (resp[order(h[, i])] - mdatay)^2)
      
            allsqresdata <- cbind(if(exists("allsqresdata")) allsqresdata
                                  else NULL,
                                  sqresdata)

            vdata <- smooth(xaxis = h[, i], yaxis = sqresdata, wts = w,
                            sp = sparams[i], spalt = spar[i])

            if(smtype == 1) {
              vdatay <- vdata$y
            }
      
            if(smtype == 3) {
              vdatay <- (fitted.values(vdata))[order(h[, i])]
            }

            sdatay <- sqrt(ifelse(vdatay <= 0, 0, vdatay))
      
            lines(plotx, mmodely - smodely, lty = 1, lwd = 1, col = "red")
            lines(plotx, mmodely + smodely, lty = 1, lwd = 1, col = "red")
            lines(plotx, mdatay - sdatay, lwd = 1, col = "blue")
            lines(plotx, mdatay + sdatay, lwd = 1, col = "blue")
      
          }

        }

        ### legend for plot ###

        if(i == num) {
          close.screen(all = T)
          par(new = TRUE, mar = c(5.1, 2.1, 1.1, 2.1))
          plot(0, 0, axes = F, xlab = "", ylab = "", type = "n",
               main = paste("Marginal model plot(s):", smoother, "smooths"))
          legend(x = 0, y = 1.05,
                 horiz = TRUE, xjust = .5, x.intersp = .5, cex = .85,
                 legend = c("Estimates for data (under F)", 
                   "Estimates for fitted values (under M)"),
                 lty = c(1, 1), lwd = 2, col = c("blue", "red"))
          par(mar = c(5.1, 4.1, 4.1, 2.1)) # don't know why this is needed
        }

      }

      close.screen(all = T)
      invisible()

      switch(menu(c("Construct Bayes marginal model plot(s) for mean",
                    "Return to command line"),
                  title = "Select one of the following") + 1,
             stop("No action\n", call. = FALSE), 
             split.screen(figs = c(round(sqrt(num), 0), ceiling(sqrt(num)))),
             stop("No action\n", call. = FALSE))
      }

      yhatsamp <- samp$yhat[included, ]

      ### cycle through number of directions h ###
    
      for(i in 1:num) {

        ### if spline, create length(x) x nsamp x 2 array ###

        if(smoother == "spline") {
          xy <- array(c(rep(h[, i], nsamp), yhatsamp), c(dim(yhatsamp), 2))
        }

        ### if loess, kernel or supsmu, create nsamp length(x) x 2 matrices ###

        if(smoother == "loess1" || smoother == "loess2" ||
           smoother == "kernel" || smoother == "supsmu") {
          xy <- vector("list", nsamp)
          for(j in 1:nsamp) {
            xy[[j]] <- cbind(yhatsamp[, j], h[, i])
          }
        }

        mmodel <- switch(smoother,
                         spline = apply(xy, 2, smooth.spline, w = w,
                           df = sparams[i], spar = spar[i]),
                         loess1 = lapply(xy, function(a)
                           loess(a[, 1] ~ a[, 2], weights = w,
                                 span = sparams[i], degree = 1,
                                 trace.hat = "approximate")),
                         loess2 = lapply(xy, function(a)
                           loess(a[, 1] ~ a[, 2], weights = w,
                                 span = sparams[i], degree = 2,
                                 trace.hat = "approximate")),
                         kernel = lapply(xy, function(a)
                           ksmooth(a[, 2], a[, 1], kernel = "normal",
                                   x.points = a[, 2], bandwidth =
                                   diff(range(a[, 2])) * sparams[i])),
                         supsmu = lapply(xy, function(a)
                           supsmu(a[, 2], a[, 1], wt = w,
                                  span = sparams[i])))

        mdata <- smooth(xaxis = h[, i], yaxis = resp, wts = w,
                        sp = sparams[i], spalt = spar[i])

        if(smtype == 1 || smtype == 2) {
          plotx <- mdata$x
          mmodely <- sapply(mmodel, function(x) x$y)
          mdatay <- mdata$y
        }
    
        if(smtype == 3) {
          plotx <- sort(h[, i])
          mmodely <- sapply(mmodel, fitted.values)[order(h[, i]),]
          mdatay <- fitted.values(mdata)[order(h[, i])]
        }

        ### when there is a restriction on range for y, eg in (0,1) ###

        if(!is.null(mrange)) {
          mmodely <- pmin(pmax(mmodely, mrange[1]), mrange[2])
          mdatay <- pmin(pmax(mdatay, mrange[1]), mrange[2])
        }

        ### construct the bmmp ###

        if(incplots) {
        screen(i)
        par(mex = 1 - (num - 1) * .05, cex = 1 - (num - 1) * .05)
#        par(mex = 1 - (num - 1) * .05, cex = 1 - (num - 1) * .05,
#            mar = c(2.4, 3.7, 1, 1))
        plot(h[, i], respj, xlim = range(h[, i]),
             ylim = c(min(resp) - mult * diff(range(resp)),
               max(resp) + mult * diff(range(resp))),
#             xlab = paste(dimnames(h)[[2]][i], " (sp:",
#               if(smoother == "spline" && spar[1] > 0) spar[i] else sparams[i],
#               ")", sep=""),
             xlab = dimnames(h)[[2]][i],
             ylab = deparse(yname), xpd = F)

        ### add lines to plot ###
    
        matlines(plotx, mmodely, lwd = 1, col = "red")
        lines(plotx, mdatay, lwd = 3, col = "blue")
#        matlines(plotx, mmodely, lwd = 1, col = "gray")
#        lines(plotx, mdatay, lwd = 2, col = "black")
        }

        if(smtype == 1 || smtype == 3) {

        ## retain sampled model mean smooths for variance smooths if needed ##

          if(incvar) {
            mmodelsamp <- switch(smoother,
                                 spline = sapply(lapply(mmodel, predict,
                                   x = h[, i]), function(x) x$y),
                                 loess1 = sapply(mmodel, fitted.values),
                                 loess2 = sapply(mmodel, fitted.values),
                                 kernel = sapply(mmodel, function(x)
                                   x$y)[order(order(h[, i])), ])
            allmmodelsamp <- rbind(if(exists("allmmodelsamp")) allmmodelsamp
                                   else NULL,
                                   mmodelsamp)
          }

          ## calculate average distance^2 from data smooth to model smooths ##

          if(bdm) {

            cat("Calculating Bayes Discrepancy Measure for",
                dimnames(h)[[2]][i], "...\n")
            mmodelsamp <- switch(smoother,
                                 spline = sapply(lapply(mmodel, predict,
                                   x = h[, i]), function(x) x$y),
                                 loess1 = sapply(mmodel, fitted.values),
                                 loess2 = sapply(mmodel, fitted.values),
                                 kernel = sapply(mmodel, function(x) x$y))
            mdatapred <- switch(smoother,
                                spline = predict(mdata, x = h[, i])$y,
                                loess1 = fitted.values(mdata),
                                loess2 = fitted.values(mdata),
                                kernel = mdata$y)
            dmd <- sum((mmodelsamp - mdatapred)^2) / (length(h[, i]) * nsamp)
            cat("Observed distance measure (DM) =", signif(dmd, 2), "\n")
            dmk <- (sum(mmodelsamp^2) - 2 *
                    apply(mmodelsamp, 2, "%*%",
                          y = apply(mmodelsamp, 1, sum)) +
                    nsamp * apply(mmodelsamp^2, 2, sum)) /
                      (length(h[, i]) * (nsamp - 1))
            cat("Model DM sample deciles:",
                format(signif(quantile(dmk, probs = seq(0, 1, 0.1)), 2)),
                fill = T)
            cat("Bayes Discrepancy Measure (mean) =",
#            cat("Bayes Discrepancy Measure for", dimnames(h)[[2]][i], "=",
                round(sum(dmk > dmd) / nsamp, 2), "\n")
#            cat(round(sum(dmk > dmd) / nsamp, 2), "")
#            bdmout[ds, i, meth] <<- round(sum(dmk > dmd) / nsamp, 2)

          }

        }

        ### legend for plot ###

        if(incplots) {
        if(i == num) {
          close.screen(all = T)
          par(new = TRUE, mar = c(5.1, 2.1, 1.1, 2.1))
#          par(new = TRUE, mar = c(0.1, 0.1, 0.1, 0.1))
          plot(0, 0, axes = F, xlab = "", ylab = "", type = "n",
               main = paste("Bayes marginal model plot(s) for mean:",
                 smoother, "smooths"))
          legend(x = 0, y = 1.05,
#          legend(x = 0, y = par("usr")[4],
                 horiz = TRUE, xjust = .5, x.intersp = .5, cex = .85,
#                 horiz = TRUE, xjust = .5, cex = .7,
                 legend = c("Estimates for data (under F)", 
                   "Estimates for fitted values (under M)"),
                 lty = c(1, 1), lwd = 2, col = c("blue", "red"))
#                 lty = c(1, 1), lwd = 2, col = c("black", "gray"))
          par(mar = c(5.1, 4.1, 4.1, 2.1)) # don't know why this is needed
        }
        }

      }

#      cat("\n")
  
      close.screen(all = T)
      invisible()

      ### calculate variance smooths if they are wanted as well ###
      ### not available for "supsmu" since it removes duplicate x's ###

      if(smtype == 1 || smtype == 3) {

        if(incvar) {

          switch(menu(c("Construct Bayes marginal model plot(s) for variance",
                        "Return to command line"),
                      title = "Select one of the following") + 1,
                 stop("No action\n", call. = FALSE),
                 split.screen(figs = c(round(sqrt(num), 0),
                                ceiling(sqrt(num)))),
                 stop("No action\n", call. = FALSE))

          vhatsamp <- matrix(rep(samp$vhat, length(w)), ncol = nsamp,
                             byrow = T) /
                               matrix(rep(w, nsamp), ncol = nsamp)
          absresdata <- sqrt(ifelse(allsqresdata <= 0, 0, allsqresdata))

          ### cycle through number of directions h ###

          for(i in 1:num) {

            ### if spline, create length(x) x nsamp x 2 arrays ###

            if(smoother == "spline") {
              xev <- array(c(rep(h[, i], nsamp), vhatsamp),
                           c(dim(yhatsamp), 2))
              xve <- array(c(rep(h[, i], nsamp),
                             (yhatsamp -
                              allmmodelsamp[((i - 1) * length(h[, i]) + 1):
                                            (i * length(h[, i])), ]) ^ 2),
                           c(dim(yhatsamp), 2))
            }

            ### if loess or kernel, create nsamp length(x) x 2 matrices ###

            if(smoother == "loess1" || smoother == "loess2" ||
               smoother == "kernel") {
              xev <- xve <- vector("list", nsamp)
              for(j in 1:nsamp) {
                xev[[j]] <- cbind(vhatsamp[, j], h[, i])
                xve[[j]] <- cbind((yhatsamp[, j] -
                                   allmmodelsamp[((i-1) * length(h[, i]) + 1):
                                                 (i * length(h[, i])), j]) ^ 2,
                                  h[, i])
              }
            }

            evmodel <- switch(smoother,
                              spline = apply(xev, 2, smooth.spline, w = w,
                                df = sparams[i], spar = spar[i]),
                              loess1 = lapply(xev, function(a)
                                loess(a[, 1] ~ a[, 2], weights = w,
                                      span = sparams[i], degree = 1,
                                      trace.hat = "approximate")),
                              loess2 = lapply(xev, function(a)
                                loess(a[, 1] ~ a[, 2], weights = w,
                                      span = sparams[i], degree = 2,
                                      trace.hat = "approximate")),
                              kernel = lapply(xev, function(a)
                                ksmooth(a[, 2], a[, 1], kernel = "normal",
                                        x.points = a[, 2], bandwidth =
                                        diff(range(a[, 2])) * sparams[i])))
      
            vemodel <- switch(smoother,
                              spline = apply(xve, 2, smooth.spline, w = w,
                                df = sparams[i], spar = spar[i]),
                              loess1 = lapply(xve, function(a)
                                loess(a[, 1] ~ a[, 2], weights = w,
                                      span = sparams[i], degree = 1,
                                      trace.hat = "approximate")),
                              loess2 = lapply(xve, function(a)
                                loess(a[, 1] ~ a[, 2], weights = w,
                                      span = sparams[i], degree = 2,
                                      trace.hat = "approximate")),
                              kernel = lapply(xve, function(a)
                                ksmooth(a[, 2], a[, 1], kernel = "normal",
                                        x.points = a[, 2], bandwidth =
                                        diff(range(a[, 2])) * sparams[i])))

            vdata <- smooth(xaxis = h[, i], yaxis = allsqresdata[, i], wts = w,
                            sp = sparams[i], spalt = spar[i])

            if(smtype == 1) {
              plotx <- vdata$x
              vmodely <- sapply(evmodel, function(x) x$y) +
                sapply(vemodel, function(x) x$y)
              vdatay <- vdata$y
            }
      
            if(smtype == 3) {
              plotx <- sort(h[, i])
              vmodely <- (sapply(evmodel, fitted.values) +
                          sapply(vemodel, fitted.values))[order(h[, i]),  ]
              vdatay <- (fitted.values(vdata))[order(h[, i])]
            }

            smodely <- sqrt(ifelse(vmodely <= 0, 0, vmodely))
            sdatay <- sqrt(ifelse(vdatay <= 0, 0, vdatay))

            ### construct the bmmp ###

            screen(i)
            par(mex = 1 - (num - 1) * .05, cex = 1 - (num - 1) * .05)
            plot(h[, i], absresdata[, i], xlim = range(h[, i]),
                 ylim = c(min(absresdata[, i]) - mult *
                   diff(range(absresdata[, i])),
                   max(absresdata[, i]) + mult * diff(range(absresdata[, i]))),
                 xlab = paste(dimnames(h)[[2]][i], " (sp:",
#                   if(smoother == "spline" && spar[1] > 0) spar[i]
                   if(smoother == "spline" && !is.null(spar)) spar[i]
                   else sparams[i],
                   ")", sep=""),
                 ylab = "|Residuals| from smoothing data", xpd = F)

            ### add lines to plot ###
    
            matlines(plotx, smodely, lwd = 1, col = "red")
            lines(plotx, sdatay, lwd = 3, col = "blue")

            # calculate average distance^2 from data smooth to model smooths #

            if(bdm) {

              cat("Calculating Bayes Discrepancy Measure for",
                  dimnames(h)[[2]][i], "...\n")
              vmodelsamp <- switch(smoother,
                                   spline = sapply(lapply(evmodel, predict,
                                     x = h[, i]), function(x) x$y) +
                                   sapply(lapply(vemodel, predict, x = h[, i]),
                                          function(x) x$y),
                                   loess1 = sapply(evmodel, fitted.values) +
                                   sapply(vemodel, fitted.values),
                                   loess2 = sapply(evmodel, fitted.values) +
                                   sapply(vemodel, fitted.values),
                                   kernel = sapply(vemodel, function(x) x$y) +
                                   sapply(evmodel, function(x) x$y))
              smodelsamp <- sqrt(ifelse(vmodelsamp <= 0, 0, vmodelsamp))
              vdatapred <- switch(smoother,
                                  spline = predict(vdata, x = h[, i])$y,
                                  loess1 = fitted.values(vdata),
                                  loess2 = fitted.values(vdata),
                                  kernel = vdata$y)
              sdatapred <- sqrt(ifelse(vdatapred <= 0, 0, vdatapred))
              dmd <- sum((smodelsamp - sdatapred)^2) / (length(h[, i]) * nsamp)
              cat("Observed distance measure (DM) =", signif(dmd, 2), "\n")
              dmk <- (sum(smodelsamp^2) - 2 *
                      apply(smodelsamp, 2, "%*%",
                            y = apply(smodelsamp, 1, sum)) +
                      nsamp * apply(smodelsamp^2, 2, sum))  /
                        (length(h[, i]) * (nsamp - 1))
              cat("Model DM sample deciles:",
                  format(signif(quantile(dmk, probs = seq(0, 1, 0.1)), 2)),
                  fill = T)
              cat("Bayes Discrepancy Measure (variance) =",
                  round(sum(dmk > dmd) / nsamp, 2), "\n")

            }

            ### legend for plot ###

            if(i == num) {
              close.screen(all = T)
              par(new = TRUE, mar = c(5.1, 2.1, 1.1, 2.1))
              plot(0, 0, axes = F, xlab = "", ylab = "", type = "n",
                   main = paste("Bayes marginal model plot(s) for variance:",
                     smoother, "smooths"))
              legend(x = 0, y = 1.05,
                     horiz = TRUE, xjust = .5, x.intersp = .5, cex = .85,
                     legend = c("Estimates for data (under F)", 
                       "Estimates for fitted values (under M)"),
                     lty = c(1, 1), lwd = 2, col = c("blue", "red"))
              par(mar = c(5.1, 4.1, 4.1, 2.1)) # don't know why this is needed

            }
            
          }

        }

      }

    }

  }

"postsamp.lm" <-
# Samples from the posterior distribution of a linear model
  function(lmobject, nsamp = 100)
  {

    rdf <- lmobject$df.residual
    x2 <- rchisq(nsamp, rdf)
    sigmasq <- deviance(lmobject) / x2
    xmatrix <- model.matrix(lmobject)
    nterms <- dim(xmatrix)[2]
    betahat <- matrix(rep(coef(lmobject), nsamp), ncol = nsamp)
    z <- matrix(rnorm(nsamp * nterms), ncol = nsamp)
    betasamp <- betahat + t(diag(sqrt(sigmasq)) %*%
                            t(backsolve(qr.R(lmobject$qr), z)))
    invisible(list(yhat = xmatrix %*% betasamp, vhat = sigmasq))

  }

"postsamp.binlog" <-
# Extract binary logistic samples from BUGS/BOA output
  function(glmobject, boaobject, nsamp = 100)
  {

    xmatrix <- model.matrix(glmobject)
    end <- dim(boaobject$chain$master[[1]])[1]
    start <- end - nsamp / 2 + 1
    betasamp <- t(rbind(boaobject$chain$master[[1]][start:end,],
                        boaobject$chain$master[[2]][start:end,]))
    dimnames(betasamp) <- list(NULL, NULL)
    invisible(list(yhat = 1 / (1 + exp(-xmatrix %*% betasamp))))

  }
